/*************************************************************************\
**  source       * ha_soliddb_backup.cc
**  directory    * mysql
**  description  * Implementation of Online Backup
**               * of solidDB Storage Engine for MySQL 5.1.
**               * 
**               * Copyright (C) 2007 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

/*
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
W  TODO LIST
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW

*** bkp_show_error()
    Implemented only for non-PLUGIN case.

*** Flush tables with read lock before doing backup.

WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
W  EOF TODO LIST
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#define INSIDE_HA_SOLIDDB_BACKUP_CC

#define _WIN32_WINNT 0x0403 /* enable TryEnterCriticalSection etc. */

#include <mysql_priv.h>
#include <mysqld_error.h>

#include <mysys_err.h>
#include <my_sys.h>
#include <my_base.h>
#include <m_ctype.h>
#include <ssmyenv.h>

#include "ha_soliddb_backup.h"

#include <mysql.h>
#if MYSQL_VERSION_ID >= 50100
#include <mysql/plugin.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#include <tab0conn.h>
#include <ssdir.h>
#ifdef __WIN__
#include <direct.h>
#endif

#ifdef __cplusplus
}
#endif

#define SAFE_DELETE(p)      { if (p) { delete(p); } }
#define SAFE_FREE(p,free)   { if (p) { free(p);   } }

/*
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
W                                                                         W
W                     S C O P E   G U A R D S                             W
W                                                                         W
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
*/
/************************************************************************\
 * The scope guard classes defined below provide a "structured" way of
 * handling "resources" which are dynamically "allocated" and "released".
 * Examples of such resources are: memory (SsMemAlloc/Free), files
 * (SsFOpen/Close), dynamically allocated structures, e.g. MYSQL
 * (mysql_init/close), etc. They ensure that a resource "allocated" in a
 * function will be "freed" when we exit from this function. The idea is
 * based on the fact that whenever we exit a function (e.g. by return or
 * throw statement) destructors for all class objects declared within the
 * function will be executed (see sg_mem description for details).
/************************************************************************/

/************************************************************************\
 * sg_mem
 *
 * This class controls dynamical memory allocation and freeing (using
 * SsMemAlloc and SsMemFree functions). The following example illustrates
 * how this works:
 *
 *            void example(...) {
 *                try {
 *                    ...
 *                    f1(...);
 *                    ...          <----------------------------------|
 *                } catch (...) {                                     |
 *                    ...          <============================|     |
 *                }                                             |     |
 *            }                                                 |     |
 *                                                              |     |
 *            void f1(...) {                                    |     |
 *                sg_mem<char*> buf;                            |     |
 *                buf.allocate(100);                            |     |
 *                ...                                           |     |
 *                f2(buf);                                      |     |
 *                ...                                           |     |
 *    |--------  return;                                        |     |
 *    |           ...                                           |     |
 *    |------->  // When exiting from f1 we are guaranteed  ----------|
 *               // that buf's destructor ~sg_mem() is          |
 *    |=======>  // executed which causes SsMemFree() to  ======|
 *    |          // be called.
 *    |       }
 *    |
 *    |       void f2(char* buf) {
 *    |           ...
 *    |=========  throw <exception>;
 *                ...
 *            }
 * NOTES
 * (1) The sg_mem class is "parameterized" by the type T of a pointer
 *     to allocated memory (NOT the type of memory elements):
 *
 *         void f(...) {
 *             sg_mem<void*> mem;
 *             sg_mem<char*> buf;
 *             ...
 *             mem.allocate(512);
 *             buf.allocate(100);
 *             ...
 *         }
 *
 * (2) The argument of the allocate() member-function is number of bytes
 *     of memory to be allocated (NOT number of elements):
 *
 *         typedef struct my_struct_t { ... } my_struct;
 *
 *         void f(...) {
 *             sg_mem<my_struct*> s;
 *             ...
 *             s.allocate(512 * sizeof(my_struct));
 *             ...
 *         }
 *
 * (3) The sg_mem class contains the T() conversion operator which
 *     allows to use class objects whenever pointer of type T is requred:
 *
 *         void f(...) {
 *             sg_mem<char*> buf;
 *             ...
 *             buf.allocate(100);
 *             ...
 *             sprintf(buf, ... );
 *             ...
 *         }
 *
 * (4) Yet one more example: when memory allocated in one function is to
 *     be freed in another function:
 *
 *         // The following function outputs some text in a memory which
 *         // is allocated in this function and which must be freed in a
 *         // calling function.
 *
 *         void get_text(sg_mem<char*>& text) {
 *             ...
 *             text.allocate(<text length>);
 *             sprintf(text, ...);
 *         }
 *
 *         void use_text() {
 *             sg_mem text;
 *             ...
 *             get_text(text);
 *             ...
 *             // text memory is freed here
 *             // (when we leave this function)
 *         }
 *
/************************************************************************/

template<typename T>
class sg_mem {
public:
        sg_mem() : m_ptr(NULL) {
        }
        ~sg_mem() {
            SAFE_FREE(m_ptr, SsMemFree);
        }
        int allocate(size_t nbytes) {
            m_ptr = (T) SsMemAlloc(nbytes);
            return m_ptr ? 0 : -1;
        }
        operator T() const {
            return m_ptr;
        }

private:
        T m_ptr;
};

/************************************************************************\
 * sg_file
 *
 * This class controls opening and closing files (using SsFOpenB/SsFOpenT
 * and SsFClose functions). The details are similar to that of the
 * sg_mem class. Example:
 *
 *     void f(...) {
 *         sg_file file;
 *         ...
 *         file.open_t("c:/temp.txt", "a");
 *         ...
 *         SsFWrite(..., file);  // Note the SS_FILE*() conversion operator
 *         ...
 *     }
 *
/************************************************************************/

class sg_file {
public:
        sg_file() : m_file(NULL) {
        }
        ~sg_file() {
            SAFE_FREE(m_file, SsFClose);
        }
        int open_b(char* path, char* flags) {
            m_file = SsFOpenB(path, flags);
            return m_file ? 0 : -1;
        }
        int open_t(char* path, char* flags) {
            m_file = SsFOpenT(path, flags);
            return m_file ? 0 : -1;
        }
        operator SS_FILE*() const {
            return m_file;
        }

private:
        SS_FILE* m_file;
};

/************************************************************************\
 * sg_dir
 *
 * This class controls opening and closing directories (using SsDirOpen
 * and SsDirClose functions). The details are similar to that of the
 * sg_file class. Example:
 *
 *     void f(...) {
 *         sg_dir dir;
 *         SsDirEntT* entry;
 *         ...
 *         dir.open("c:/work");
 *         ...
 *         entry = SsDirRead(dir);  // Note the SsDirT*() conversion operator
 *         ...
 *     }
 *
/************************************************************************/

class sg_dir {
public:
        sg_dir() : m_dir(NULL) {
        }
        ~sg_dir() {
            SAFE_FREE(m_dir, SsDirClose);
        }
        int open(char* path) {
            m_dir = SsDirOpen(path);
            return m_dir ? 0 : -1;
        }
        operator SsDirT*() const {
            return m_dir;
        }

private:
        SsDirT* m_dir;
};

/************************************************************************\
 * sg_mysql
 *
 * This class controls creating and freeing MYSQL objects (using mysql_init
 * and mysql_close functions). The details are similar to that of other
 * sg_xxx classes (see example in the sg_mysql_res class description).
/************************************************************************/

class sg_mysql {
public:
        sg_mysql() : m_ptr(NULL) {
        }
        ~sg_mysql() {
            SAFE_FREE(m_ptr, mysql_close);
        }
        int init(MYSQL* mysql = NULL) {
            m_ptr = mysql_init(mysql);
            return m_ptr ? 0 : -1;
        }
        operator MYSQL*() const {
            return m_ptr;
        }

private:
        MYSQL* m_ptr;
};

/************************************************************************\
 * sg_mysql_res
 *
 * This class controls creating and freeing MYSQL_RES objects (using
 * mysql_store_result andmysql_free_result functions). The details are
 * similar to that of other scope_guard classes. Example:
 *
 *     void f(...) {
 *         sg_mysql mysql;
 *         char* query = "SHOW DATABASES";
 *         sg_mysql_res query_result;
 *         MYSQL_ROW  row;
 *         ...
 *         mysql.init();
 *         ...
 *         mysql_real_query(mysql, query, strlen(query));
 *         ...
 *         query_result.store(mysql);
 *         ...
 *         row = mysql_fetch_row(query_result);
 *         ...
 *     }
 *
/************************************************************************/

class sg_mysql_res {
public:
        sg_mysql_res() : m_ptr(NULL) {
        }
        ~sg_mysql_res() {
            SAFE_FREE(m_ptr, mysql_free_result);
        }
        int store(MYSQL* mysql) {
            m_ptr = mysql_store_result(mysql);
            return m_ptr ? 0 : -1;
        }
        operator MYSQL_RES*() const {
            return m_ptr;
        }

private:
        MYSQL_RES* m_ptr;
};

/*
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
W                                                                         W
W                     D D L   M O N I T O R I N G                         W
W                                                                         W
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
*/
#define SS_SEMNUM_SOLIDDB_DDL_MONITOR (SS_SEMNUM_SS_LIB)

typedef struct ddl_monitoring
{
    unsigned int  number_of_backups;
    unsigned int  number_of_ddls;
    SsFlatMutexT  mutex;
    SsMesT*       ddl_finish_event;
} ddl_monitoring_t;

static ddl_monitoring_t ddl_monitor;

/*#***********************************************************************\
 *            init_ddl_monitor()
 *
 * Initialize DDL monitor
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    ddl_monitoring_t  ddl_monitor
 */
void init_ddl_monitor()
{
        ddl_monitor.number_of_backups = 0;
        ddl_monitor.number_of_ddls = 0;
        SsFlatMutexInit(&ddl_monitor.mutex, SS_SEMNUM_SOLIDDB_DDL_MONITOR);
        ddl_monitor.ddl_finish_event = SsMesCreateLocal();
        SsMesSend(ddl_monitor.ddl_finish_event);
}

/*#***********************************************************************\
 *            free_ddl_monitor()
 *
 * Free DDL monitor
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    ddl_monitoring_t  ddl_monitor
 */
void free_ddl_monitor()
{
        SsFlatMutexDone(ddl_monitor.mutex);
        SsMesFree(ddl_monitor.ddl_finish_event);
}

/*#***********************************************************************\
 *            block_solid_DDL
 *
 * Wait until no DDL commands are active and add one more backup which
 * blocks DDL for Solid tables
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    ddl_monitoring_t  ddl_monitor
 */
void block_solid_DDL()
{
        bool ddl_active = false;

        SsFlatMutexLock(ddl_monitor.mutex);

        if (ddl_monitor.number_of_backups < UINT_MAX) {
            ddl_monitor.number_of_backups++;
        }

        ddl_active = ddl_monitor.number_of_ddls > 0;

        SsFlatMutexUnlock(ddl_monitor.mutex);

        if (ddl_active) {
            SsMesWait(ddl_monitor.ddl_finish_event);
        }
}

/*#***********************************************************************\
 *              unblock_solid_DDL
 *
 * Remove one backup which blocks DDL for Solid tables
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    ddl_monitoring_t  ddl_monitor
 */
void unblock_solid_DDL()
{
        SsFlatMutexLock(ddl_monitor.mutex);

        if (ddl_monitor.number_of_backups) {
            ddl_monitor.number_of_backups--;
        }

        SsFlatMutexUnlock(ddl_monitor.mutex);
}

/*#***********************************************************************\
 *              is_solid_DDL_blocked
 *
 * Check if there is any backup which blocks DDL for Solid tables
 *
 * Parameters
 *    none
 *
 * Return value
 *    bool  - TRUE if there is at least one backup which blocks DDL
 *            for Solid tables
 *
 * Globals used
 *    ddl_monitoring_t  ddl_monitor
*/
bool is_solid_DDL_blocked()
{
        bool ddl_blocked = false;

        SsFlatMutexLock(ddl_monitor.mutex );
        ddl_blocked = ddl_monitor.number_of_backups > 0;
        SsFlatMutexUnlock(ddl_monitor.mutex);

        return ddl_blocked;
}

/*#***********************************************************************\
 *              start_solid_DDL
 *
 * Start DDL and register this within monitor if DDL execution currently
 * is not blocked.
 *
 * Parameters
 *    rc    - int [out]
 *
 * Return value
 *    bool  - TRUE if DDL can be executed (i.e. if DDL execution is not
 *            blocked currently). In this case finish_solid_DDL() must be
 *            called after DDL execution (otherwise Backup will never be
 *            able to start).
 *
 * Globals used
 *    ddl_monitoring_t  ddl_monitor
 */
bool start_solid_DDL(int &rc)
{
        bool ddl_blocked = false;

        SsFlatMutexLock(ddl_monitor.mutex);

        ddl_blocked = ddl_monitor.number_of_backups > 0;

        if (!ddl_blocked) {
            if (ddl_monitor.number_of_ddls < UINT_MAX) {
                ddl_monitor.number_of_ddls++;
            }
            SsMesReset(ddl_monitor.ddl_finish_event);
        }

        SsFlatMutexUnlock(ddl_monitor.mutex);

        if (ddl_blocked) {
            rc = HA_ERR_WRONG_COMMAND;
            sql_print_error("DDL commands are temporarily disabled, "
                            "possibly because on-line backup is in progress. "
                            "Try again later.\n");

            /* ui_msg_message(BACKUP_MSG_FAILED_S, su_err_geterrstr(*p_errh)); */

            ss_pprintf(("DDL commands are temporarily disabled, "
                        "possibly because on-line backup is in progress. "
                        "Try again later.\n"));
        }

        return (!ddl_blocked);
}

/*#***********************************************************************\
 *              finish_solid_DDL
 *
 * Finish DDL execution. Must be called in pair with each start_solid_DDL.
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    ddl_monitoring_t  ddl_monitor
 */
void finish_solid_DDL()
{
        unsigned int prev_number_of_ddls = 0;

        SsFlatMutexLock(ddl_monitor.mutex);
        prev_number_of_ddls = ddl_monitor.number_of_ddls;

        if (ddl_monitor.number_of_ddls) {
            ddl_monitor.number_of_ddls--;
        }

        if (prev_number_of_ddls && !ddl_monitor.number_of_ddls) {
            /* this DDL was the last one, now there is no DDL active */
            SsMesSend(ddl_monitor.ddl_finish_event);
        }

        SsFlatMutexUnlock(ddl_monitor.mutex);
}

/*
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
W                                                                         W
W                   B A C K U P   D I R E C T O R Y                       W
W                                                                         W
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
*/
#define SS_SEMNUM_SOLIDDB_BACKUP_DIR  (SS_SEMNUM_SS_LIB)

SsFlatMutexT  backupdir_mutex;
char          soliddb_backupdir_buf[FN_REFLEN] = "";
char*         soliddb_backupdir = soliddb_backupdir_buf;

/*#***********************************************************************\
 *            init_backup_directory
 *
 * Initialize backup directory using solid configuration
 *
 * Parameters
 *     inifile  su_inifile_t* inifile [in, out]
 *
 * Return value
 *    bool      TRUE if successfull
 *
 * Globals used
 *
 */
bool init_backup_directory(su_inifile_t* inifile)
{
        bool res = TRUE;

        if (soliddb_backupdir == soliddb_backupdir_buf) {
            return TRUE;
        }

        if (soliddb_backupdir != NULL && soliddb_backupdir[0]) {
            strncpy(soliddb_backupdir_buf, soliddb_backupdir, sizeof(soliddb_backupdir_buf) - 1);
        }

        soliddb_backupdir = soliddb_backupdir_buf;

        ss_pprintf_1(("BackupDir=%s\n", soliddb_backupdir));
        su_inifile_putstring(inifile, (char*) SU_DBE_GENERALSECTION,
                            (char*) SU_DBE_BACKUPDIR, soliddb_backupdir);

        return TRUE;
}

/*#***********************************************************************\
 *            soliddb_check_backupdir
 *
 * Check directory for backup files.
 *
 * Parameters :
 *
 *     THD*      thd, in, use
 *     set_var*  var, in, use, variable value from MySQL
 *
 * Return value : 0
 *
 * Globals used :
 */
int soliddb_check_backupdir(
        MYSQL_THD                thd,
#if MYSQL_VERSION_ID >= 50100
        struct st_mysql_sys_var* var,
        void*                    save,
        struct st_mysql_value*   value
#else
        set_var*                 var
#endif
        )
{
        const char *str;
#if MYSQL_VERSION_ID >= 50100
        char buff[STRING_BUFFER_USUAL_SIZE];
        int length;

        length= sizeof(buff);
        str= value->val_str(value, buff, &length);
#else
        str = var ? var->value->str_value.c_ptr_quick() : NULL;
#endif

        ss_pprintf_1(("BackupDir= [%s]\n", str ? str : "NULL"));

        return str ? 0 : -1;
}

/*#***********************************************************************\
 *            solid_update_backupdir
 *
 * Update directory for backup files.
 *
 * Parameters :
 *
 *     THD*      thd, in, use
 *     set_var*  var, in, use
 *
 * Return value : TRUE if successfull
 *                FALSE otherwice
 *
 * Globals used :
 */

#if MYSQL_VERSION_ID >= 50100
void soliddb_update_backupdir(
        MYSQL_THD thd,
        struct st_mysql_sys_var* var,
        void*                    save,
        void*                    value)
#else
bool soliddb_update_backupdir(
        MYSQL_THD thd,
        set_var*                 var)
#endif
{
        char* new_value;
#if MYSQL_VERSION_ID >= 50100
        char buff[STRING_BUFFER_USUAL_SIZE];
        int length;

        length= sizeof(buff);
        new_value = (char *)value;

        if (!new_value) {
            new_value = (char *)"";
        }
#else

        new_value = (char *)var->value->str_value.c_ptr_quick();

        if (var->value->null_value || !new_value) {
            new_value = (char *)"";
        }
#endif

        DBUG_ENTER("soliddb_update_backupdir");
        ss_pprintf_1(("BackupDir=%s\n", new_value));

        new_value = SsStrTrimLeft(new_value);
        SsStrTrimRight(new_value);

        SsFlatMutexLock(backupdir_mutex);
        strcpy(soliddb_backupdir_buf, new_value);
        SsFlatMutexUnlock(backupdir_mutex);

#if MYSQL_VERSION_ID >= 50100
        DBUG_VOID_RETURN;
#else
        DBUG_RETURN(0);
#endif
}

/*
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
W                                                                         W
W                     O N L I N E   B A C K U P                           W
W                                                                         W
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
*/
#define SS_SEMNUM_SOLIDDB_BACKUP  (SS_SEMNUM_SS_LIB)

#if MYSQL_VERSION_ID >= 50100
#define TABLE_TYPE  enum legacy_db_type
#else
#define TABLE_TYPE  enum db_type
#endif

#define TABLE_TYPE_UNKNOWN  DB_TYPE_UNKNOWN
#define TABLE_TYPE_MYISAM   DB_TYPE_MYISAM
#if MYSQL_VERSION_ID >= 50100
#define TABLE_TYPE_SOLID    DB_TYPE_SOLID
#else
#define TABLE_TYPE_SOLID    DB_TYPE_SOLID_DB
#endif

#define MYSQL_EXT_FRM   "frm"
#define MYSQL_EXT_MYD   "MYD"
#define MYSQL_EXT_MYI   "MYI"
#define MYSQL_EXT_TRG   "TRG"
#define MYSQL_EXT_TRN   "TRN"

#if MYSQL_VERSION_ID >= 50100
#define THD_KILLED(thd) (thd_killed(thd))
#else
#define THD_KILLED(thd) (thd->killed)
#endif

/************************************************************************\
 *  Typedefs
/************************************************************************/

typedef struct bkp_table_st   bkp_table_t;  // table to be backed up
typedef struct bkp_dbase_st   bkp_dbase_t;  // database to be backed up
typedef struct bkp_ctx_st     bkp_ctx_t;    // backup context
typedef struct bkp_params_st  bkp_params_t; // backup parameters

/************************************************************************\
 *  Global Variables
/************************************************************************/

bkp_ctx_t*  bkp_ctx                 = NULL;   // Current backup context
bool        soliddb_backup_emptydir = false;  // By default backup directory should not be emptied
bool        soliddb_backup_myisam   = false;  // By default MyISAM tables should not be backed up
bool        soliddb_backup_system   = false;  // By default system tables should not be backed up
bool        soliddb_backup_config   = false;  // By default config file should not be backed up

/************************************************************************\
 *  Forward Declarations
/************************************************************************/

static char*        bkp_parse_path(char** text);
static int          bkp_parse_onoff(char** text, ulong* opt);
static void         bkp_parse_params(char* text, bkp_params_t* data);

static  void        init_bkp_ctx(bkp_ctx_t*);
static  bkp_ctx_t*  create_bkp_ctx();
static  void        free_bkp_ctx(bkp_ctx_t*);

static su_ret_t     bkp_do_abort();
static su_ret_t     bkp_do_wait();
static su_ret_t     bkp_do_start(rs_sysi_t*, bkp_params_t*, su_err_t**);

static void         bkp_set_start_options(bkp_params_t*);
static su_ret_t     bkp_set_directory(bkp_params_t*);
static bool         bkp_dir_is_empty(char* path);
static void         bkp_dir_make_empty(char* path);

static su_ret_t     bkp_prepare();
static su_ret_t     bkp_set_list_of_databases(MYSQL*);
static su_ret_t     bkp_set_list_of_tables(MYSQL*, bkp_dbase_t*);
static int          bkp_list_of_tables_query(sg_mem<char*>& query, const char* dbasename);
static su_ret_t     bkp_set_list_of_triggers(MYSQL*, bkp_table_t*);
static int          bkp_list_of_triggers_query(sg_mem<char*>& query, const char* dbasename, const char* tablename);
static su_ret_t     bkp_set_table_size(bkp_table_t*);

void SS_CALLBACK    bkp_thread_fn();
static bool         bkp_aborted();
static su_ret_t     bkp_dbase(bkp_dbase_t*);
static su_ret_t     bkp_table(bkp_table_t*);
static su_ret_t     bkp_soliddb_table(bkp_table_t*);
static su_ret_t     bkp_myisam_table(bkp_table_t*);
static su_ret_t     bkp_triggers(bkp_table_t*);
static su_ret_t     bkp_file(const char* dbase, const char* name, const char* extension);
static su_ret_t     bkp_soliddb();
static su_ret_t     bkp_createcheckpoint();

#ifndef MYSQL_DYNAMIC_PLUGIN
static su_ret_t     bkp_do_status();
static void         bkp_show_status();
#endif

/************************************************************************\
 *  Backup Log
/************************************************************************/

static void         bkp_log_begin();
static void         bkp_log_end();
static void         bkp_log_dbase_begin(bkp_dbase_t*);
static void         bkp_log_dbase_end(bkp_dbase_t*);
static void         bkp_log_table_begin(bkp_table_t*);
static void         bkp_log_table_end(bkp_table_t*);
static void         bkp_log_trigger_begin(const char* name);
static void         bkp_log_trigger_end(const char* name);
static void         bkp_log_file_begin(const char* name, const char* extension);
static void         bkp_log_file_end(const char* name, const char* extension);
static void         bkp_log_text(const char*);
static char*        bkp_log_datetime(char* buff, int size);
static int          bkp_log_write(const char* text);

/*#***********************************************************************\
 *
 *            bkp_show_error
 *
 * Show backup error
 *
 * Parameters
 *    errmsg        char* [in] Error message to show
 *    system_error  bool [in]  TRUE means internal (system) error
 *
 * Return value
 *    none
 *
 * Globals used
 *    none
 */
static void bkp_show_error(const char* errmsg, bool system_error = false)
{
#ifndef MYSQL_DYNAMIC_PLUGIN
        MYSQL_THD   thd = current_thd;
        Protocol*   protocol = thd->protocol;
        List<Item>  field_list;
        Item*       item;

        field_list.push_back(item = new Item_empty_string("status", NAME_LEN * 2));
        item->maybe_null = 1;
        field_list.push_back(item = new Item_empty_string("description", NAME_LEN * 2));
        item->maybe_null = 1;
        protocol->send_fields(&field_list, Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF);

        protocol->prepare_for_resend();
        protocol->store(system_error ? "SYSTEM ERROR" : "ERROR", system_charset_info);
        protocol->store(errmsg, system_charset_info);
        protocol->write();
        send_eof(thd);
#else
        // TODO
#endif
}

/*#***********************************************************************\
 *            init_backup
 *
 * Initialize backup
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
void init_backup()
{
        SsFlatMutexInit(&backupdir_mutex, SS_SEMNUM_SOLIDDB_BACKUP_DIR);
        bkp_ctx = create_bkp_ctx();
}

/*#***********************************************************************\
 *            free_backup
 *
 * Free Backup
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *
 *    bkp_ctx_t*    bkp_ctx
 *    SsFlatMutexT  soliddb_backup_mutex
 */
void free_backup()
{
        free_bkp_ctx(bkp_ctx);
        SsFlatMutexDone(backupdir_mutex);
}

/************************************************************************\
 *  Backup Syntax
 *
 *  SET GLOBAL SOLIDDB_ADMIN_COMMAND = 'BACKUP parameters'
 *
 *  parameters :: start | abort | wait | status { if not plugin }
 *
 *  start  :: [-DIR=path]
 *            [-EMPTYDIR[=ON|OFF]]
 *            [-MYISAM[=ON|OFF]]
 *            [-SYSTEM[=ON|OFF]]
 *            [-CONFIG[=ON|OFF]]
 *  abort  :: -ABORT
 *  wait   :: -WAIT
 *  status :: -STATUS   // if not plugin
/************************************************************************/

typedef enum {
        BKP_CMD_UNDEF,
        BKP_CMD_START,
        BKP_CMD_ABORT,
        BKP_CMD_WAIT,
#ifndef MYSQL_DYNAMIC_PLUGIN
        BKP_CMD_STATUS
#endif
} bkp_cmd_t;

#define BKP_OPT_UNDEF   0x0000
#define BKP_OPT_OFF     0x0001
#define BKP_OPT_ON      0x0002

struct bkp_params_st {
        bkp_cmd_t cmd;          // backup command
        char*     dir;          // backup directory
        ulong     opt_dempty;   // -EMPTYDIR option
        ulong     opt_myisam;   // -MYISAM option
        ulong     opt_system;   // -SYSTEM option
        ulong     opt_config;   // -CONFIG option
};

/*#***********************************************************************\
 *            bkp_parse_params
 *
 * Parse backup parameters
 *
 * Parameters
 *    text    char* [in] String containing parameters to parse
 *    data    bkp_params_t* [out] Note. In case of invalid parameters
 *            data->cmd = BKP_CMD_UNDEF.
 *
 * Return value
 *    none
 *
 * Globals used
 *    none
 */
static void bkp_parse_params(char* text, bkp_params_t* data)
{
        bkp_cmd_t cmd;

        text = text ? text : (char*)"";

        data->cmd = BKP_CMD_UNDEF;
        data->dir = NULL;
        data->opt_myisam = BKP_OPT_UNDEF;
        data->opt_system = BKP_OPT_UNDEF;
        data->opt_config = BKP_OPT_UNDEF;
        data->opt_dempty = BKP_OPT_UNDEF;

        while (true) {
            if (!*(text = SsStrTrimLeft(text))) {
                break;
            }
            if (*text != '-') {
                bkp_show_error("Backup parameters must start with '-'");
                goto ERROR;
            }

            if (!SsStrnicmp(text, "-abort", 6)) {
                text += 6;
                cmd = BKP_CMD_ABORT;

#ifndef MYSQL_DYNAMIC_PLUGIN
            } else if (!SsStrnicmp(text, "-status", 7)) {
                text += 7;
                cmd = BKP_CMD_STATUS;
#endif

            } else if (!SsStrnicmp(text, "-wait", 5)) {
                text += 5;
                cmd = BKP_CMD_WAIT;

            } else if (!SsStrnicmp(text, "-dir", 4)) {
                text += 4;
                cmd = BKP_CMD_START;
                if (!(data->dir = bkp_parse_path(&text))) {
                    goto ERROR;
                }

            } else if (!SsStrnicmp(text, "-myisam", 7)) {
                text += 7;
                cmd = BKP_CMD_START;
                if (bkp_parse_onoff(&text, &data->opt_myisam)) {
                    goto ERROR;
                }

            } else if (!SsStrnicmp(text, "-system", 7)) {
                text += 7;
                cmd = BKP_CMD_START;
                if (bkp_parse_onoff(&text, &data->opt_system)) {
                    goto ERROR;
                }

            } else if (!SsStrnicmp(text, "-config", 7)) {
                text += 7;
                cmd = BKP_CMD_START;
                if (bkp_parse_onoff(&text, &data->opt_config)) {
                    goto ERROR;
                }

            } else if (!SsStrnicmp(text, "-emptydir", 9)) {
                text += 9;
                cmd = BKP_CMD_START;
                if (bkp_parse_onoff(&text, &data->opt_dempty)) {
                    goto ERROR;
                }

            } else {
                bkp_show_error("Unknown backup parameter");
                goto ERROR;
            }

            if (data->cmd == BKP_CMD_UNDEF) {
                data->cmd = cmd;
            } else if (data->cmd != cmd) {
                bkp_show_error("Incompatible backup parameters");
                goto ERROR;
            }
        }

        // Empty parameters can be only in case of START
        if (data->cmd == BKP_CMD_UNDEF) {
            data->cmd = BKP_CMD_START;
        }

        // Nothing should be left in the text
        if (*(text = SsStrTrimLeft(text))) {
            bkp_show_error("Incorrect backup parameters");
            goto ERROR;
        }

        return;

ERROR:
        data->cmd = BKP_CMD_UNDEF;
        return;
}

/*#***********************************************************************\
 *            bkp_parse_path
 *
 * Parse the '=path' part of the backup parameters. Used only in the
 * bkp_parse_params() function.
 *
 * Parameters
 *    text      char** [in, out] On entry points at the beginning of the
 *              '=path' part. On exit points at the next char following
 *              the '=path' part
 *
 * Return value
 *    pointer to the 'path', or NULL in case of error
 *
 * Globals used
 *    none
 */
static char* bkp_parse_path(char** text)
{
        char* ptr = SsStrTrimLeft(*text);
        char* path = NULL;
        bool stroke;
        bool quotat;

        if (*ptr != '=') {
            bkp_show_error("Missed '=' in -dir parameter");
            return NULL;
        }
        ptr = SsStrTrimLeft(++ptr);

        stroke = (*ptr == '\'');
        quotat = (*ptr == '"');
        if (stroke || quotat) {
            ptr = SsStrTrimLeft(++ptr);
        }

        path = ptr;

        for (path = ptr; *ptr; ptr++) {
          if (stroke && *ptr == '\'' ||
              quotat && *ptr == '"'  ||
              !stroke && !quotat && *ptr == ' ') {
              break;
          }
          if (*ptr == '\\' &&
              (*(ptr+1) == '\'' || *(ptr+1) == '"' || *(ptr+1) == ' ')) {
              ptr++;
          }
        }

        if (*ptr) {
            *ptr++ = '\0';
        }

        *text = ptr;
        return path;
}

/*#***********************************************************************\
 *            bkp_parse_onoff
 *
 * Parse an optional ['=ON|OFF'] part of the backup parameters. Used oonly
 * in the bkp_parse_params() function.
 *
 * Parameters
 *    text    char** [in, out] On entry points at the beginning of the
 *            ['=ON|OFF'] part. On exit points at the next char following
 *            the ['=ON|OFF'] part
 *    opt     ulong* [out] On exit is assigned to BKP_OPT_ON|OFF. In case
 *            of error is not changed
 *
 * Return value
 *    non-zero value in case of error
 *
 * Globals used
 *    none
 */
static int bkp_parse_onoff(char** text, ulong* opt)
{
        char* ptr = *text;
        int res = 0;

        if (*(ptr = SsStrTrimLeft(ptr)) == '=') {
            ptr += 1;
            ptr = SsStrTrimLeft(ptr);
            if  (!SsStrnicmp(ptr, "ON", 2)) {
                ptr += 2;
                *opt = BKP_OPT_ON;
            } else if (!SsStrnicmp(ptr, "OFF", 3)) {
                ptr += 3;
                *opt = BKP_OPT_OFF;
            } else {
                bkp_show_error("Incorrect value of boolean parameter");
                res = -1;
            }
        } else {
            *opt = BKP_OPT_ON;
        }

        *text = ptr;
        return res;
}

/************************************************************************
 *  clist_t class
 *
 *  We need to have several types of lists in the backup:
 *  - list of strings which are names of triggers defined for a table
 *  - list of tables to be backed up
 *  - list of databases to be backed up
 *  clist_t is implemented as a wrapper of the su_list_t type. Note that
 *  additionally to itemdel() function (used in su_list_t) we also use
 *  itemnew() function for creating new item instances which are added
 *  to the list.
 ************************************************************************/

// Position of an item in the list is a pointer to this item
typedef su_list_node_t* position_t;

class clist_t
{
public:
        static void* operator new(size_t size);
        static void operator delete(void* ptr);

        clist_t(void* (*itemnew)(void*) = NULL, void (*itemdel)(void*) = NULL);
        ~clist_t();

        inline  position_t  add(void* item);
        inline  void        del(position_t);
        inline  position_t  del_and_next(position_t);
        inline  position_t  first();
        inline  position_t  next(position_t);
        inline  bool        is_empty();
        inline  void        clear();

private:
        su_list_t*  m_list;
        void*       (*m_itemnew)(void*);
};

void* clist_t::operator new(size_t size) {
        return (void*) SsMemAlloc(size);
}

void clist_t::operator delete(void* ptr) {
        SsMemFree(ptr);
}

/************************************************************************
 *  Constructor
 */
clist_t::clist_t(void* (*itemnew)(void*), void (*itemdel)(void*)) {
        m_list = su_list_init(itemdel);
        m_itemnew = itemnew;
}

/************************************************************************
 *  Destructor
 */
clist_t::~clist_t() {
        su_list_done(m_list);
}

/************************************************************************
 *  Add new item to the list and return its position.
 */
position_t clist_t::add(void* item) {
        if (m_itemnew) {
            item = m_itemnew(item);
        }
        return su_list_insertlast(m_list, item);
}

/************************************************************************
 *  Delete list item at a given position.
 */
void clist_t::del(position_t pos) {
        su_list_remove(m_list, pos);
}

/************************************************************************
 *  Delete list item at a given position and return position of the
 * next item.
 */
position_t clist_t::del_and_next(position_t pos) {
        return su_list_removeandnext(m_list, pos);
}

/************************************************************************
 *  Return position of the first list item.
 */
position_t clist_t::first() {
        return su_list_first(m_list);
}

/************************************************************************
 *  Return position of the next list item.
 */
position_t clist_t::next(position_t pos) {
        return su_list_next(m_list, pos);
}

/************************************************************************
 *  Is list empty?
 */
bool clist_t::is_empty() {
        return m_list->list_length == 0;
}

/************************************************************************
 *  Make list empty.
 */
void clist_t::clear() {
        su_list_clear(m_list);
}

/************************************************************************
 *  List of strings
 *
 *  List of strings is clist_t* object created by the new_string_list()
 *  function. E.g.
 *
 *      clist_t* list = new_string_list();
 *
 *  Implementationally, it uses two private (internal) functions
 *  stringnew() and stringdel() for creating and deleting string list
 *  items (see clist_t::clist_t constructor and clist_t::add() function).
 *  E.g.
 *
 *      list->add("string");
 *
 *  adds new string "string" to list.
 *  The get_string_at() function returns pointer to the string with given
 *  position in the list. E.g.
 *
 *      position_t pos;
 *      for (pos = list->first(); pos; pos = list->next(pos)) {
 *          char* str = get_string_at(pos);
 *          ...
 *      }
 ************************************************************************/

/************************************************************************
 *  Internal function used to create new string item.
 *  See new_string_list() function.
 */
static void* stringnew(void* p) {
        ss_assert(p);
        return SsMemStrdup((char*) p);
}

/************************************************************************
 *  Internal function used to delete string item.
 *  See new_string_list() function.
 */
static void stringdel(void* p) {
        SAFE_FREE(p, SsMemFree);
}

/************************************************************************
 *  Create new list of strings.
 */
inline clist_t* new_string_list() {
        return new clist_t(stringnew, stringdel);
}

/************************************************************************
 *  Return pointer to the string at a given position in the list.
 */
inline char* get_string_at(position_t pos) {
        return (char*) pos->ln_data;
}

/************************************************************************
 *  Table to be backed up
 ************************************************************************/

struct bkp_table_st {
        bkp_dbase_t*  dbase;      // database to which the table belongs
        char*         name;       // table name
        TABLE_TYPE    type;       // table type: TALE_TYPE_MYISAM, TABLE_TYPE_SOLID
        double        size;       // table size
        clist_t*      triggers;   // list of triggers defined for the table
};

/************************************************************************
 *  List of tables
 *
 *  List of tables is clist_t* object created by the new_table_list()
 *  function. E.g.
 *
 *      clist_t* list = new_table_list();
 *
 *  Implementationally, it uses two private (internal) functions
 *  tablenew() and tabledel() for creating and deleting table list items
 *  (see clist_t::clist_t constructor and clist_t::add() function). E.g.
 *
 *      list->add("table");
 *
 *  adds new table with name "table" to the list.
 *  The get_table_at() function returns pointer to the table with given
 *  position in the list. E.g.
 *
 *      position_t pos;
 *      for (pos = list->first(); pos; pos = list->next(pos)) {
 *          bkp_table_t* table = get_table_at(pos);
 *          ...
 *      }
 ************************************************************************/

/************************************************************************
 *  Internal function used to create new table item.
 *  See new_table_list() function.
 */
static void* tablenew(void* tablename)
{
        ss_assert(tablename);
        bkp_table_t* table = (bkp_table_t*) SsMemAlloc(sizeof(bkp_table_t));
        table->dbase  = NULL;
        table->name   = (char*) SsMemStrdup((char*) tablename);
        table->type   = TABLE_TYPE_UNKNOWN;
        table->size   = 0;
        table->triggers = new_string_list();
        return table;
}

/************************************************************************
 *  Internal function used to delete table item.
 *  See new_table_list() function.
 */
static void tabledel(void* p)
{
        if (p) {
            bkp_table_t* table = (bkp_table_t*) p;
            SAFE_FREE(table->name, SsMemFree);
            SAFE_DELETE(table->triggers);
            SsMemFree(table);
        }
}

/************************************************************************
 *  Create new list of tables.
 */
inline clist_t* new_table_list() {
        return new clist_t(tablenew, tabledel);
}

/************************************************************************
 *  Return pointer to the table at a given position in the list.
 */
inline bkp_table_t* get_table_at(position_t pos) {
        return (bkp_table_t*) pos->ln_data;
}

/************************************************************************
 *  Database to be backed up
 ************************************************************************/

struct bkp_dbase_st {
        char*     name;     // database name
        double    size;     // database size
        clist_t*  tables;   // list of tables
};

/************************************************************************
 *  List of databases
 *
 *  List of databases is clist_t* object created by the new_dbase_list()
 *  function. E.g.
 *
 *      clist_t* list = new_dbase_list();
 *
 *  Implementationally, it uses two private (internal) functions
 *  dbasenew() and dbasedel() for creating and deleting database list
 *  items (see clist_t::clist_t constructor and clist_t::add() function).
 *  E.g.
 *
 *      list->add("test_database");
 *
 *  adds new database with name "test_database" to the list.
 *  The get_dbase_at() function returns pointer to the database with given
 *  position in the list. E.g.
 *
 *      position_t pos;
 *      for (pos = list->first(); pos; pos = list->next(pos)) {
 *          bkp_dbase_t* dbase = get_dbase_at(pos);
 *          ...
 *      }
 ************************************************************************/

/************************************************************************
 *  Internal function used to create new database item.
 *  See new_dbase_list() function.
 */
static void* dbasenew(void* dbasename)
{
        ss_assert(dbasename);
        bkp_dbase_t* dbase = (bkp_dbase_t*) SsMemAlloc(sizeof(bkp_dbase_t));
        dbase->name = (char*) SsMemStrdup((char*) dbasename);
        dbase->size = 0;
        dbase->tables = new_table_list();
        return dbase;
}

/************************************************************************
 *  Internal function used to delete database item.
 *  See new_dbase_list() function.
 */
static void dbasedel(void* p)
{
        if (p) {
            bkp_dbase_t* dbase = (bkp_dbase_t*) p;
            SAFE_FREE(dbase->name, SsMemFree);
            SAFE_DELETE(dbase->tables);
            SsMemFree(dbase);
        }
}

/************************************************************************
 *  Create new list of database.
 */
inline clist_t* new_dbase_list() {
        return new clist_t(dbasenew, dbasedel);
}

/************************************************************************
 *  Return pointer to the database at a given position in the list.
 */
inline bkp_dbase_t* get_dbase_at(position_t pos) {
        return (bkp_dbase_t*) pos->ln_data;
}

/************************************************************************
 *  Backup context
 ************************************************************************/

struct bkp_ctx_st {
        SsFlatMutexT  mutex;          // Mutex to access shared resources
        bool          active;         // backup is active (is in progress)
        bool          failed;         // backup failed
        bool          abort;          // backup is (to be) aborted
        bool          killed;         // backup thread is killed
        char          dir[FN_REFLEN]; // backup destination directory
        ulong         opt_myisam;     // are myisam tables to be backed up?
        ulong         opt_system;     // are system tables to be backed up?
        ulong         opt_config;     // is option (config) file to be backed up?
        ulong         opt_dempty;     // is backup directory be emptied?
        double        total_size;     // size of all files to be backed up
        double        done_size;      // size of all backed up files
        clist_t*      databases;      // databases involved in the backup
        MYSQL_THD     thd;            // thd within which backup was activated
        SS_FILE*      flog;           // log file
        rs_sysi_t*    cd;
        su_err_t**    p_errh;
};

/*#***********************************************************************\
 *            create_bkp_ctx
 *
 * Create new backup context instance.
 *
 * Parameters
 *    none
 *
 * Return value
 *    bkp_ctx_t*  created backup context instance (NULL if failed)
 *
 * Globals used
 *    none
 */
static bkp_ctx_t* create_bkp_ctx()
{
        bkp_ctx_t*  ctx = (bkp_ctx_t*) SsMemAlloc(sizeof(bkp_ctx_t));
        clist_t* dblist = new_dbase_list();

        if (!ctx) {
            bkp_show_error("create_bkp_ctx(): SsMemAlloc failed", true);
            return NULL;
        }

        SsFlatMutexInit(&ctx->mutex, SS_SEMNUM_SOLIDDB_BACKUP);
        ctx->databases = dblist;

        if (!dblist) {
            SsFlatMutexDone(ctx->mutex);
            SAFE_FREE(ctx, SsMemFree);
            bkp_show_error("create_bkp_ctx(): new_dbase_list failed", true);
            return NULL;
        }

        init_bkp_ctx(ctx);

        return ctx;
}

/*#***********************************************************************\
 *            init_bkp_ctx
 *
 * Initialize backup context. Should be called before starting new backup
 * process.
 *
 * Parameters
 *    ctx     bkp_ctx_t* [out]  Backup context to be initialized.
 *
 * Return value
 *    none
 *
 * Globals used
 *    none
 */
static void init_bkp_ctx(bkp_ctx_t* ctx)
{
        ctx->active     = FALSE;
        ctx->failed     = FALSE;
        ctx->abort      = FALSE;
        ctx->killed     = FALSE;
        ctx->dir[0]     ='\0';
        ctx->opt_myisam = 0;
        ctx->opt_system = 0;
        ctx->opt_config = 0;
        ctx->opt_dempty = 0;
        ctx->total_size = 0;
        ctx->done_size  = 0;

        ctx->databases->clear();

        ctx->thd    = NULL;
        ctx->flog   = NULL;
        ctx->cd     = NULL;
        ctx->p_errh = NULL;
}

/*#***********************************************************************\
 *            free_bkp_ctx
 *
 * Free backup context.
 *
 * Parameters
 *    ctx     bkp_ctx_t* [out]  Backup context to be freed.
 *
 * Return value
 *    none
 *
 * Globals used
 *    none
 */
static void free_bkp_ctx(bkp_ctx_t* ctx)
{
        if (ctx) {
            SsFlatMutexDone(ctx->mutex);
            SAFE_DELETE(ctx->databases);
            SAFE_FREE(ctx->flog, SsFClose);
            SsMemFree(ctx);
        }
}

/*#***********************************************************************\
 *            do_backup
 *
 * Execute solidDB 'BACKUP <parameters>' command.
 *
 * Parameters
 *    cd        rs_sysi_t* [in]
 *    parmstr   char*      [in]
 *    p_errh    su_err_t** [in out]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
su_ret_t do_backup(rs_sysi_t* cd, char* parmstr, su_err_t** p_errh)
{
        bkp_params_t parmdata;
        su_ret_t rc = SU_SUCCESS;

        if (!bkp_ctx) {
            bkp_show_error("Backup context is not created", true);
            rc = DBE_ERR_FAILED;
            goto FINISH;
        }

        bkp_parse_params(parmstr, &parmdata);

        switch (parmdata.cmd) {
        case BKP_CMD_ABORT:
            rc = bkp_do_abort();
            break;
#ifndef MYSQL_DYNAMIC_PLUGIN
        case BKP_CMD_STATUS:
            rc = bkp_do_status();
            break;
#endif
        case BKP_CMD_WAIT:
            rc = bkp_do_wait();
            break;
        case BKP_CMD_START:
            rc = bkp_do_start(cd, &parmdata, p_errh);
            break;
        case BKP_CMD_UNDEF:
        default:
            rc = DBE_ERR_FAILED;
            break;
        }


FINISH:
        return rc;
}

/*#***********************************************************************\
 *            bkp_do_abort
 *
 * Execute solidDB 'BACKUP -ABORT' command
 *
 * Parameters
 *    none
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_do_abort()
{
        bool active;

        SsFlatMutexLock(bkp_ctx->mutex);
        if ((active = bkp_ctx->active)) {
            bkp_ctx->abort = true;
        }
        SsFlatMutexUnlock(bkp_ctx->mutex);

        while (active) {
            SsThrSleep(20);
            SsFlatMutexLock(bkp_ctx->mutex);
            active = bkp_ctx->active;
            SsFlatMutexUnlock(bkp_ctx->mutex);
        }

#ifndef MYSQL_DYNAMIC_PLUGIN
        bkp_show_status();
#endif
        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_do_wait
 *
 * Execute solidDB 'BACKUP -WAIT' command
 *
 * Parameters
 *    none
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_do_wait()
{
        bool active = true;

        while (active) {
            SsFlatMutexLock(bkp_ctx->mutex);
            active = bkp_ctx->active;
            SsFlatMutexUnlock(bkp_ctx->mutex);
            SsThrSleep(20);
        }

#ifndef MYSQL_DYNAMIC_PLUGIN
        bkp_show_status();
#endif
        return SU_SUCCESS;
}

#ifndef MYSQL_DYNAMIC_PLUGIN
/*#***********************************************************************\
 *            bkp_do_status
 *
 * Execute solidDB 'BACKUP -STATUS' command
 *
 * Parameters
 *    none
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_do_status()
{
        bkp_show_status();
        return SU_SUCCESS;
}
#endif

#ifndef MYSQL_DYNAMIC_PLUGIN
/*#***********************************************************************\
 *            bkp_show_status
 *
 * Show current status of the backup
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_show_status()
{
        MYSQL_THD   thd = current_thd;
        Protocol*   protocol = thd->protocol;
        List<Item>  field_list;
        Item*       item;
        char*       status;
        double      total_size;
        double      done_size;
        uint32      done;
        char*       directory;
        char*       opt_myisam;
        char*       opt_system;
        char*       opt_config;
        char*       opt_dempty;

        // Hack to avoid loosing a connection (see also
        // sql_parse.cpp: mysql_execute_command(): case SQLCOM_SET_OPTION)
        thd->net.no_send_ok = 1;

        SsFlatMutexLock(bkp_ctx->mutex);
        status  = bkp_ctx->killed ? (char*)"KILLED"  :
                  bkp_ctx->abort  ? (char*)"ABORTED" :
                  bkp_ctx->failed ? (char*)"FAILED" :
                  bkp_ctx->active ? (char*)"ACTIVE"  :
                                    (char*)"FINISHED";
        total_size  = bkp_ctx->total_size;
        done_size   = bkp_ctx->done_size;
        directory   = bkp_ctx->dir;
        opt_myisam  = bkp_ctx->opt_myisam ? (char*)"ON" : (char*)"OFF";
        opt_system  = bkp_ctx->opt_system ? (char*)"ON" : (char*)"OFF";
        opt_config  = bkp_ctx->opt_config ? (char*)"ON" : (char*)"OFF";
        opt_dempty  = bkp_ctx->opt_dempty ? (char*)"ON" : (char*)"OFF";
        SsFlatMutexUnlock(bkp_ctx->mutex);

        done = (uint32) ((total_size ? done_size / total_size : 0) * 100);

        field_list.push_back(item = new Item_empty_string("status", NAME_LEN * 2));
        item->maybe_null = 1;
        field_list.push_back(new Item_return_int("total size", 7, FIELD_TYPE_LONGLONG));
        item->maybe_null = 1;
        field_list.push_back(new Item_return_int("done size", 7, FIELD_TYPE_LONGLONG));
        item->maybe_null = 1;
        field_list.push_back(new Item_return_int("percent complete", 7, FIELD_TYPE_LONG));
        item->maybe_null = 1;
        field_list.push_back(item = new Item_empty_string("directory", NAME_LEN * 2));
        item->maybe_null = 1;
        field_list.push_back(item = new Item_empty_string("empty dir", NAME_LEN * 2));
        item->maybe_null = 1;
        field_list.push_back(item = new Item_empty_string("myisam", NAME_LEN * 2));
        item->maybe_null = 1;
        field_list.push_back(item = new Item_empty_string("system", NAME_LEN * 2));
        item->maybe_null = 1;
        protocol->send_fields(&field_list, Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF);

        protocol->prepare_for_resend();
        protocol->store(status, system_charset_info);
        protocol->store((ulonglong) total_size);
        protocol->store((ulonglong) done_size);
        protocol->store(done);
        protocol->store(directory, system_charset_info);
        protocol->store(opt_dempty, system_charset_info);
        protocol->store(opt_myisam, system_charset_info);
        protocol->store(opt_system, system_charset_info);
        protocol->write();
        send_eof(thd);
}
#endif

/*#***********************************************************************\
 *            bkp_do_start
 *
 * Execute solidDB 'BACKUP <start>' command
 *
 * Parameters
 *    cd        rs_sysi_t*  [in]
 *    parms     bkp_params_t* [in]
 *    err       su_err_t**  [in]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_do_start(rs_sysi_t* cd, bkp_params_t* parms, su_err_t** err)
{
        SsThreadT* bkp_thread;
        bool ddl_blocked = FALSE;
        su_ret_t rc = SU_SUCCESS;

        // Check whether backup is active or not currently
        SsFlatMutexLock(bkp_ctx->mutex);
        if (bkp_ctx->active) {
            SsFlatMutexUnlock(bkp_ctx->mutex);
            bkp_show_error("Backup can not be started because it is active currently");
            return DBE_ERR_FAILED;
        }
        SsFlatMutexUnlock(bkp_ctx->mutex);

        // Initialize backup context
        init_bkp_ctx(bkp_ctx);
        SsFlatMutexLock(bkp_ctx->mutex);
        bkp_ctx->active = TRUE;
        SsFlatMutexUnlock(bkp_ctx->mutex);
        bkp_ctx->thd = current_thd;
        bkp_ctx->cd = cd;
        bkp_ctx->p_errh = err;

        // Set options (-system, -myisam, etc)
        bkp_set_start_options(parms);

        // Set backup directory
        if ((rc = bkp_set_directory(parms)) != SU_SUCCESS) {
            su_err_init_text(err, DBE_ERR_FAILED, "Invalid backup directory");
            goto FAILURE;
        }

        // Block DDL execution during backup
        block_solid_DDL();

        // Prepare context of the new backup
        if ((rc = bkp_prepare()) != SU_SUCCESS) {
            goto FAILURE;
        }

        // Create and run backup thread
        bkp_thread = SsThrInit(bkp_thread_fn,
          "backup", SS_THREAD_USEDEFAULTSTACKSIZE);

        if (!SsThrEnableBool(bkp_thread)) {
            SsThrDone(bkp_thread);
            rc = DBE_ERR_FAILED;
            goto FAILURE;
        }

#ifndef MYSQL_DYNAMIC_PLUGIN
        bkp_show_status();
#endif
        return SU_SUCCESS;

FAILURE:
        if (ddl_blocked) {
            unblock_solid_DDL();
        }
        SsFlatMutexLock(bkp_ctx->mutex);
        bkp_ctx->active = FALSE;
        bkp_ctx->failed = TRUE;
        SsFlatMutexUnlock(bkp_ctx->mutex);
        return rc;
}

/*#***********************************************************************\
 *            bkp_set_start_options
 *
 *    This function sets backup option values in the bkp_ctx structure
 *    using option values in the backup start command and default option
 *    values.
 *
 * Parameters :
 *    parms     bkp_params_t* [in]
 *
 * Return value :
 *    none
 *
 * Globals used :
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_set_start_options(bkp_params_t* parms)
{
        SsFlatMutexLock(bkp_ctx->mutex);
        if (parms->opt_myisam) {
            bkp_ctx->opt_myisam = parms->opt_myisam >> 1;
        } else {
            bkp_ctx->opt_myisam = soliddb_backup_myisam;
        }
        if (parms->opt_system) {
            bkp_ctx->opt_system = parms->opt_system >> 1;
        } else {
            bkp_ctx->opt_system = soliddb_backup_system;
        }
        if (parms->opt_config) {
            bkp_ctx->opt_config = parms->opt_config >> 1;
        } else {
            bkp_ctx->opt_config = soliddb_backup_config;
        }
        if (parms->opt_dempty) {
            bkp_ctx->opt_dempty = parms->opt_dempty >> 1;
        } else {
            bkp_ctx->opt_dempty = soliddb_backup_emptydir;
        }
        SsFlatMutexUnlock(bkp_ctx->mutex);
}

/*#***********************************************************************\
 *            bkp_set_directory
 *
 * This function sets backup directory in the bkp_ctx structure
 * using optional directory spec in the backup start command and
 * the default backup directory spec.
 *
 * Parameters
 *    parms     bkp_params_t* [in]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_set_directory(bkp_params_t* parms)
{
        dbe_db_t* soliddb = (dbe_db_t*)rs_sysi_db(bkp_ctx->cd);

        if (parms->dir && *parms->dir) {
            // Use directory defined in the backup command
            SsFlatMutexLock(bkp_ctx->mutex);
            strcpy(bkp_ctx->dir, parms->dir);
            SsFlatMutexUnlock(bkp_ctx->mutex);

         } else if (soliddb_backupdir && *soliddb_backupdir) {
            // Use default backup directory
            SsFlatMutexLock(backupdir_mutex);
            SsFlatMutexLock(bkp_ctx->mutex);
            strcpy(bkp_ctx->dir, soliddb_backupdir);
            SsFlatMutexUnlock(bkp_ctx->mutex);
            SsFlatMutexUnlock(backupdir_mutex);

        } else {
            // Nor implicit neither default backup directory is defined
            bkp_show_error("Backup directory is not defined");
            return DBE_ERR_FAILED;
        }

        // Check that backup directory really exists
        if (dbe_db_backupcheck(soliddb, bkp_ctx->dir, bkp_ctx->p_errh) != DBE_RC_SUCC) {
            bkp_show_error(su_err_geterrstr(*bkp_ctx->p_errh));
            return DBE_ERR_FAILED;
        }

        if (bkp_ctx->opt_dempty) {
            // Backup directory is to be emptied
            bkp_dir_make_empty(bkp_ctx->dir);
            if (!bkp_dir_is_empty(bkp_ctx->dir)) {
                bkp_show_error("Failed to empty the backup directory");
                return DBE_ERR_FAILED;
            }
        } else {
            // Backup directory must already be empty
            if (!bkp_dir_is_empty(bkp_ctx->dir)) {
                bkp_show_error("Backup directory is not empty");
                return DBE_ERR_FAILED;
            }
        }

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_dir_is_empty
 *
 * Check whether directory is empty.
 *
 * Parameters
 *    path      char* [in] Directory path
 *
 * Return value
 *    bool      TRUE if directory is empty. FALSE otherwise or if
 *              directory does not exists.
 *
 * Globals used
 *    none
 */
static bool bkp_dir_is_empty(char* path)
{
        sg_dir dir;
        SsDirEntT* dir_entry;

        if (dir.open(path)) {
            return FALSE;
        }

        while ((dir_entry = SsDirRead(dir))) {
            if (SsStrcmp(dir_entry->filename, ".") && SsStrcmp(dir_entry->filename, "..")) {
                break;
            }
        }

        return (dir_entry ? FALSE : TRUE);
}

/*#***********************************************************************\
 *            bkp_dir_make_empty
 *
 * Make directory empty.
 *
 * Parameters
 *    path      char* [in]
 *
 * Return value
 *    none
 *
 * Globals used
 *    none
 */
static void bkp_dir_make_empty(char* path)
{
        sg_dir dir;
        SsDirEntT* dir_entry;
        char subpath[512];

        if (dir.open(path)) {
            return;
        }

        while ((dir_entry = SsDirRead(dir))) {
            if (!SsStrcmp(dir_entry->filename, ".") || !SsStrcmp(dir_entry->filename, "..")) {
                continue;
            }

            sprintf(subpath, "%s/%s", path, dir_entry->filename);
            if (SsDExist(subpath)) {
                bkp_dir_make_empty(subpath);
                rmdir(subpath);
            } else {
                SsFRemove(subpath);
            }
        }
}

/*#***********************************************************************\
 *            bkp_prepare
 *
 * Prepare context for new backup process
 *
 * Parameters
 *    none
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_prepare()
{
        dbe_db_t* solid_db = (dbe_db_t*) rs_sysi_db(bkp_ctx->cd);
        sg_mysql mysql;
        clist_t* dbases;
        clist_t* tables;
        su_ret_t rc = SU_SUCCESS;

        if (mysql.init()) {
            bkp_show_error("Failed to create MYSQL object", TRUE);
            return DBE_ERR_FAILED;
        }
        if (mysql_options(mysql, MYSQL_READ_DEFAULT_GROUP, "mysqld")) {
            bkp_show_error("Failed to set up connection options for MySQL server");
            return DBE_ERR_FAILED;
        }
        if (!mysql_real_connect(mysql, NULL, "root", NULL, NULL, 0, NULL, 0)) {
            bkp_show_error("Failed to connect to MySQL server");
            return DBE_ERR_FAILED;
        }

        // Set list of all databases (only database name and no other data)
        if ((rc = bkp_set_list_of_databases(mysql)) != SU_SUCCESS) {
            return rc;
        }

        dbases = bkp_ctx->databases;

        // Look through all databases
        for (position_t posdbase = dbases->first(); posdbase; ) {
            bkp_dbase_t* dbase = get_dbase_at(posdbase);

            // Exclude I_S database from backup
            if (!SsStricmp(dbase->name, "information_schema")) {
                posdbase = dbases->del_and_next(posdbase);
                continue;
            }
            // Exclude system database from backup if it is not required
            if (!SsStricmp(dbase->name, "mysql") && !bkp_ctx->opt_system) {
                posdbase = dbases->del_and_next(posdbase);
                continue;
            }
            // Set list of all tables (only table name and its type)
            if ((rc = bkp_set_list_of_tables(mysql, dbase)) != SU_SUCCESS) {
                return rc;
            }

            tables = dbase->tables;

            // Look through all tables
            for (position_t postable = tables->first(); postable; ) {
                bkp_table_t* table = get_table_at(postable);

                // Exclude tables with unknown (not solid or myisam) type from backup
                if (table->type == TABLE_TYPE_UNKNOWN) {
                    postable = tables->del_and_next(postable);
                    continue;
                }
                // Exclude non-system myisam tables from backup if it is not required
                if (table->type == TABLE_TYPE_MYISAM &&
                    SsStricmp(dbase->name, "mysql") &&
                    !bkp_ctx->opt_myisam) {
                    postable= tables->del_and_next(postable);
                    continue;
                }
                // Set list of all triggers associated with the table
                if ((rc = bkp_set_list_of_triggers(mysql, table)) != SU_SUCCESS) {
                    return rc;
                }
                // Set size of (all files associated with) the table
                if ((rc = bkp_set_table_size(table)) != SU_SUCCESS) {
                    return rc;
                }
                // Exclude tables with zero size of its files
                if (table->size == 0) {
                    postable = tables->del_and_next(postable);
                    continue;
                }
                // Accumulate total size of the database
                dbase->size += table->size;
                postable = tables->next(postable);
            }

            // Exclude databases with zero size
            if (dbase->size == 0) {
                posdbase = dbases->del_and_next(posdbase);
                continue;
            }
            // Accumulate total size of files to be backed up
            bkp_ctx->total_size += dbase->size;
            posdbase = dbases->next(posdbase);
        }

        // Add size of the soliddb files
        bkp_ctx->total_size += dbe_db_getdbsize(solid_db) * 1024 +
                               dbe_db_getlogsize(solid_db) * 1024;

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_set_list_of_databases
 *
 * This function creates list of all databases in the backup context.
 * On exit, each database in the list contains only database name and no
 * other data (which is to be filled later on).
 *
 * Parameters
 *    mysql     MYSQL* [in]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_set_list_of_databases(MYSQL* mysql)
{
        char* query = "SHOW DATABASES";
        sg_mysql_res query_result;
        MYSQL_ROW row;

        if (mysql_real_query(mysql, query, strlen(query))) {
            return DBE_ERR_FAILED;
        }
        if (query_result.store(mysql)) {
            return DBE_ERR_FAILED;
        }

        while ((row = mysql_fetch_row(query_result))) {
            if (!(bkp_ctx->databases->add((char*)row[0]))) {
                return DBE_ERR_FAILED;
            }
        }

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_set_list_of_tables
 *
 * This function creates list of all tables in a given database.
 * On exit, each table in the list contains only table name, table type,
 * and reference to the database to which it belongs. All other table
 * data is to be filled later on.
 *
 * Parameters
 *    mysql     MYSQL* [in]
 *    dbase     bkp_dbase_t* [in, out]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_set_list_of_tables(MYSQL* mysql, bkp_dbase_t* dbase)
{
        sg_mem<char*> query;
        sg_mysql_res  query_result;
        MYSQL_ROW  row;
        position_t pos;
        bkp_table_t* table;

        if (bkp_list_of_tables_query(query, dbase->name)) {
            return DBE_ERR_FAILED;
        }
        if (mysql_real_query(mysql, query, strlen(query))) {
            return DBE_ERR_FAILED;
        }
        if (query_result.store(mysql)) {
            return DBE_ERR_FAILED;
        }

        while ((row = mysql_fetch_row(query_result))) {
            if (!(pos = dbase->tables->add((char*)row[0]))) {
                return DBE_ERR_FAILED;
            }
            table = get_table_at(pos);
            table->dbase = dbase;
            table->type  = !SsStricmp((char*)row[1], "soliddb") ? TABLE_TYPE_SOLID :
                           !SsStricmp((char*)row[1], "myisam")  ? TABLE_TYPE_MYISAM :
                           TABLE_TYPE_UNKNOWN;
        }

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_list_of_tables_query
 *
 * Create SQL query to select all tables in the database
 *
 * Parameters
 *    query       sg_mem<char*> [out]
 *                guarded memory buffer to write the query
 *    dbasename   const char* [in]
 *
 * Return value
 *    int         zero on success, non-zero on failure
 *
 * Globals used
 *    none
 */
static int bkp_list_of_tables_query(sg_mem<char*>& query, const char* dbasename)
{
        char* format = "SELECT table_name, engine"
                       " FROM information_schema.tables"
                       " WHERE table_schema = '%s'";

        if (query.allocate(strlen(format) + strlen(dbasename) + 1)) {
            return -1;
        }

        sprintf(query, format, dbasename);
        return 0;
}

/*#***********************************************************************\
 *            bkp_set_list_of_triggers
 *
 * This function creates list of all trigger names associated with a given
 * table.
 *
 * Parameters
 *    mysql     MYSQL* [in]
 *    table     bkp_table_t* [in, out]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_set_list_of_triggers(MYSQL* mysql, bkp_table_t* table)
{
        sg_mem<char*> query;
        sg_mysql_res  query_result;
        MYSQL_ROW  row;

        if (bkp_list_of_triggers_query(query, table->dbase->name, table->name)) {
            return DBE_ERR_FAILED;
        }
        if (mysql_real_query(mysql, query, strlen(query))) {
            return DBE_ERR_FAILED;
        }
        if (query_result.store(mysql)) {
            return DBE_ERR_FAILED;
        }

        while ((row = mysql_fetch_row(query_result))) {
            if (SsStricmp((char*)row[1], table->dbase->name)) {
                // TODO: Trigger and table belong to different databases
            }
            if (!(table->triggers->add((char*)row[0]))) {
                return DBE_ERR_FAILED;
            }
        }

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_list_of_triggers_query
 *
 * Create SQL query to select all triggers associate with the table
 *
 * Parameters
 *    query       sg_mem<char*> [out]
 *                guarded memory buffer to write the query
 *    dbasename   const char* [in]
 *    tablename   const char* [in]
 *
 * Return value
 *    int         zero on success, non-zero on failure
 *
 * Globals used
 *    none
 */
static int bkp_list_of_triggers_query(sg_mem<char*>& query, const char* dbasename, const char* tablename)
{
        char* format = "SELECT trigger_name, trigger_schema"
                       " FROM information_schema.triggers"
                       " WHERE event_object_schema = '%s'"
                       "   AND event_object_table = '%s'";
        
        if (query.allocate(strlen(format) + strlen(dbasename) + strlen(tablename) + 1)) {
            return -1;
        }

        sprintf(query, format, dbasename, tablename);
        return 0;
}

/*#***********************************************************************\
 *            bkp_set_table_size
 *
 * This function calculates total size of all files associated with a
 * given table and saves it in the backup tables structure.
 *
 * Parameters
 *    table     bkp_table_t* [in, out]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    none
 */
static su_ret_t bkp_set_table_size(bkp_table_t* table)
{
        char path[512];

        sprintf(path, "%s/%s/%s.%s", mysql_data_home,
            table->dbase->name, table->name, MYSQL_EXT_FRM);
        table->size = SsFSize(path);

        if (table->type == TABLE_TYPE_MYISAM) {
            sprintf(path, "%s/%s/%s.%s", mysql_data_home,
                table->dbase->name, table->name, MYSQL_EXT_MYD);
            table->size += SsFSize(path);
            sprintf(path, "%s/%s/%s.%s", mysql_data_home,
                table->dbase->name, table->name, MYSQL_EXT_MYI);
            table->size += SsFSize(path);
        }

        if (!table->triggers->is_empty()) {
            sprintf(path, "%s/%s/%s.%s", mysql_data_home,
                table->dbase->name, table->name, MYSQL_EXT_TRG);
            table->size += SsFSize(path);
        }

        for (position_t pos = table->triggers->first(); pos; pos = table->triggers->next(pos)) {
            sprintf(path, "%s/%s/%s.%s", mysql_data_home,
                table->dbase->name, get_string_at(pos), MYSQL_EXT_TRN);
            table->size += SsFSize(path);
        }

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_aborted
 *
 * Check whether current backup process should be aborted for some
 * external reason.
 *
 * Parameters
 *    none
 *
 * Return value
 *    bool        TRUE if backup process shold be aborted
 *
 * Globals used
 *    none
 */
static bool bkp_aborted()
{
        bool aborted;

        SsFlatMutexLock(bkp_ctx->mutex);
        bkp_ctx->killed = THD_KILLED(bkp_ctx->thd);
        aborted = bkp_ctx->abort || bkp_ctx->killed;
        SsFlatMutexUnlock(bkp_ctx->mutex);

        return aborted;
}

#define RETURN_IF_ABORTED \
        if (bkp_aborted()) { return DBE_ERR_BACKUP_ABORTED; }

#define DBE_ERR_CONTINUE SU_SUCCESS

/*#***********************************************************************\
 *            bkp_thread_fn
 *
 * The backup thread function backs up all stuff (databases, tables,
 * triggers) information about which is gathered in the global backup
 * context variable bkp_ctx.
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
void SS_CALLBACK bkp_thread_fn()
{
        position_t pos;

        bkp_log_begin();

        for (pos = bkp_ctx->databases->first(); pos; pos = bkp_ctx->databases->next(pos)) {
            if (bkp_dbase(get_dbase_at(pos)) != SU_SUCCESS) {
                goto FINISH;
            }
        }

        if (bkp_soliddb() != SU_SUCCESS) {
            goto FINISH;
        }


FINISH:
        unblock_solid_DDL();
        bkp_log_end();
        SsFlatMutexLock(bkp_ctx->mutex);
        bkp_ctx->active = FALSE;
        SsFlatMutexUnlock(bkp_ctx->mutex);
}

/*#***********************************************************************\
 *            bkp_dbase
 *
 * Back up all files associated with a given database
 *
 * Parameters
 *    dbase     bkp_dbase_t* [in]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_dbase(bkp_dbase_t* dbase)
{
        char path[512];
        position_t pos;
        su_ret_t rc;

        RETURN_IF_ABORTED;
        bkp_log_dbase_begin(dbase);

        sprintf(path, "%s/%s", bkp_ctx->dir, dbase->name);

        RETURN_IF_ABORTED;
        if (!SsMkdir(path)) {
            bkp_log_text(" ... ERROR: Failed to create database backup directory");
            return DBE_ERR_CONTINUE; // nevertheless continue
        }

        for (pos = dbase->tables->first(); pos; pos = dbase->tables->next(pos)) {
            if ((rc = bkp_table(get_table_at(pos))) != SU_SUCCESS) {
                return rc;
            }
        }

        bkp_log_dbase_end(dbase);
        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_table
 *
 * Back up all files associated with a given table
 *
 * Parameters
 *    table     bkp_table_t* [in]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_table(bkp_table_t* table)
{
        su_ret_t rc = SU_SUCCESS;

        RETURN_IF_ABORTED;
        bkp_log_table_begin(table);

        switch (table->type) {
        case TABLE_TYPE_SOLID:
            rc = bkp_soliddb_table(table);
            break;
        case TABLE_TYPE_MYISAM:
            rc = bkp_myisam_table(table);
            break;
        }
        if (rc != SU_SUCCESS) {
            return rc;
        }

        if ((rc = bkp_triggers(table)) != SU_SUCCESS) {
            return rc;
        }

        SsFlatMutexLock(bkp_ctx->mutex);
        bkp_ctx->done_size += table->size;
        SsFlatMutexUnlock(bkp_ctx->mutex);

        bkp_log_table_end(table);
        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_soliddb_table
 *
 * Back up .FRM file associated with a given solid table
 *
 * Parameters
 *    table     bkp_table_t* [in]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_soliddb_table(bkp_table_t* table)
{
        su_ret_t rc;

        if ((rc = bkp_file(table->dbase->name, table->name, MYSQL_EXT_FRM)) != SU_SUCCESS) {
            return rc;
        }

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_myisam_table
 *
 * Back up .FRM, .NYD, and MYI files associated with a given myisam table
 *
 * Parameters
 *    table     bkp_table_t* [in]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_myisam_table(bkp_table_t* table)
{
        su_ret_t rc;
        //TODO: bkp_lock_table(table->dbase->name, table->name)

        if ((rc = bkp_file(table->dbase->name, table->name, MYSQL_EXT_FRM)) != SU_SUCCESS) {
            return rc;
        }
        if ((rc = bkp_file(table->dbase->name, table->name, MYSQL_EXT_MYD)) != SU_SUCCESS) {
            return rc;
        }
        if ((rc = bkp_file(table->dbase->name, table->name, MYSQL_EXT_MYI)) != SU_SUCCESS) {
            return rc;
        }

        //TODO: bkp_unlock_tables()
        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_triggers
 *
 * Back up .TRG and .TRN trigger files associated with a given table
 *
 * Parameters
 *    table     bkp_table_t* [in]
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_triggers(bkp_table_t* table)
{
        position_t pos;
        su_ret_t rc;

        if (table->triggers->is_empty()) {
            return SU_SUCCESS;
        }

        if ((rc = bkp_file(table->dbase->name, table->name, MYSQL_EXT_TRG)) != SU_SUCCESS) {
            return rc;
        }

        for (pos = table->triggers->first(); pos; pos = table->triggers->next(pos)) {
            char* trigname = get_string_at(pos);

            RETURN_IF_ABORTED;
            bkp_log_trigger_begin(trigname);

            if ((rc = bkp_file(table->dbase->name, trigname, MYSQL_EXT_TRN)) != SU_SUCCESS) {
                return rc;
            }

            bkp_log_trigger_end(trigname);
        }

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_file
 *
 * Back up database file defined by its name and extension
 *
 * Parameters
 *    dbase     const char* [in]  Database name
 *    name      const char* [in]  File name
 *    extension const char* [in]  File extension (without leading dot)
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_file(const char* dbase, const char* name, const char* extension)
{
        char src_path[512];
        char dst_path[512];
        sg_file src_file;
        sg_file dst_file;
        char buffer[2048];
        int  nread;
        int  nwrite;

        RETURN_IF_ABORTED;
        bkp_log_file_begin(name, extension);

        sprintf(src_path, "%s/%s/%s.%s", mysql_data_home,
            dbase, name, extension);
        sprintf(dst_path, "%s/%s/%s.%s", bkp_ctx->dir,
            dbase, name, extension);

        RETURN_IF_ABORTED;
        if (src_file.open_b(src_path, "r")) {
            bkp_log_text(" ... ERROR: Failed to open source file");
            return DBE_ERR_CONTINUE; // nevertheless continue
        }

        RETURN_IF_ABORTED;
        if (dst_file.open_b(dst_path, "a")) {
            bkp_log_text(" ... ERROR: Failed to create destination file");
            return DBE_ERR_CONTINUE; // nevertheless continue
        }

        while ((nread = SsFRead(buffer, 1, sizeof(buffer), src_file))) {
            RETURN_IF_ABORTED;
            if ((nwrite = SsFWrite(buffer, 1, nread, dst_file)) < nread) {
                bkp_log_text(" ... ERROR: Failed to write to destination file");
                return DBE_ERR_CONTINUE; // nevertheless continue
            }
        }

        bkp_log_text(" ... OK");
        bkp_log_file_end(name, extension);

        return SU_SUCCESS;
}

/*#***********************************************************************\
 *            bkp_soliddb
 *
 * Back up soliddb files.
 *
 * Parameters
 *    none
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_soliddb()
{
        dbe_db_t* soliddb = (dbe_db_t*) rs_sysi_db(bkp_ctx->cd);
        double  before_size;
        double  file_size;
        double  count_size;
        long    block_size;
        su_ret_t rc;

        RETURN_IF_ABORTED;
        before_size = bkp_ctx->done_size;
        file_size   = ((double) dbe_db_getdbsize(soliddb)) * 1024 +
                      ((double) dbe_db_getlogsize(soliddb)) * 1024;
        dbe_cfg_getbackup_blocksize(soliddb->db_go->go_cfg, &block_size);
        count_size = 0;

        if ((rc = bkp_createcheckpoint()) != SU_SUCCESS) {
            return rc;
        }

        RETURN_IF_ABORTED;
        if ((rc = dbe_db_backupstart(soliddb, bkp_ctx->dir, FALSE, bkp_ctx->p_errh)) != DBE_RC_SUCC) {
            return rc;
        }

        while (TRUE) {
            RETURN_IF_ABORTED;
            if ((rc = dbe_db_backupadvance(soliddb, bkp_ctx->p_errh)) != DBE_RC_CONT) {
                break;
            }
            if ((count_size += (double) block_size) > file_size) {
                count_size = file_size;
            }
            SsFlatMutexLock(bkp_ctx->mutex);
            if ((bkp_ctx->done_size = before_size + count_size) > bkp_ctx->total_size) {
                bkp_ctx->done_size = bkp_ctx->total_size;
            }
            SsFlatMutexUnlock(bkp_ctx->mutex);
        }

        dbe_db_backupstop(soliddb);

        if (rc == DBE_RC_END && count_size < file_size) {
            SsFlatMutexLock(bkp_ctx->mutex);
            if ((bkp_ctx->done_size = before_size + file_size) > bkp_ctx->total_size) {
                bkp_ctx->done_size = bkp_ctx->total_size;
            }
            SsFlatMutexUnlock(bkp_ctx->mutex);
        }

FINISH:
        return (rc == DBE_RC_END) ? SU_SUCCESS : DBE_ERR_FAILED;
}

/*#***********************************************************************\
 *            bkp_createcheckpoint
 *
 * 
 *
 * Parameters
 *    none
 *
 * Return value
 *    solidDB return code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static su_ret_t bkp_createcheckpoint()
{
        tb_connect_t* tc;
        su_ret_t rc;

        RETURN_IF_ABORTED;
        tc = tb_sysconnect_init((tb_database_t*)rs_sysi_tabdb(bkp_ctx->cd));

        if ((rc = tb_createcheckpoint(tc, TRUE)) != DBE_RC_SUCC) {
            su_err_init(bkp_ctx->p_errh, rc);
        }
        tb_sysconnect_done(tc);

        return (rc == DBE_RC_SUCC) ? SU_SUCCESS : DBE_ERR_FAILED;
}

/************************************************************************\
 *  Backup Log
/************************************************************************/

#define BACKUP_LOG_FILE     "backup.log"
#define BKP_MAX_LOG_INDENT  3
char*   bkp_log_indent[4] = {"","  ","    ","      "};
uint    bkp_log_curr_indent;

/*#***********************************************************************\
 *            bkp_log_begin
 *
 * Open log file
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_begin()
{
        char buff[1024];
        char time[32];

        bkp_log_curr_indent = 0;

        sprintf(buff, "%s\\%s", bkp_ctx->dir, BACKUP_LOG_FILE);
        if (!(bkp_ctx->flog = SsFOpenT(buff, "a+"))) {
            goto FAILURE;
        }

        sprintf(buff, "\nBACKUP STARTED"
                      "\n* date-time:  %s"
                      "\n* directory:  %s"
                      "\n* empty dir:  %s"
                      "\n* myisam:     %s"
                      "\n* system:     %s"
                      "\n* total size: %i bytes"
                      "\n",
                      bkp_log_datetime(time, 32),
                      bkp_ctx->dir,
                      bkp_ctx->opt_dempty ? "ON" : "OFF",
                      bkp_ctx->opt_myisam ? "ON" : "OFF",
                      bkp_ctx->opt_system ? "ON" : "OFF",
                      (int) bkp_ctx->total_size);

        if (bkp_log_write(buff)) {
            goto FAILURE;
        }
        if (SsFFlush(bkp_ctx->flog)) {
            goto FAILURE;
        }

        return;

FAILURE:
        bkp_ctx->flog = NULL;
}

/*#***********************************************************************\
 *            bkp_log_end
 *
 * Close log file
 *
 * Parameters
 *    none
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_end()
{
        char buff[1024];
        char time[32];

        if (!bkp_ctx->flog) {
            return;
        }

        sprintf(buff, "\nBACKUP FINISHED"
                      "\n* date-time: %s"
                      "\n",
                      bkp_log_datetime(time, 32));

        if (bkp_log_write(buff)) {
            goto FINISH;
        }
        if (SsFFlush(bkp_ctx->flog)) {
            goto FINISH;
        }

FINISH:
        SAFE_FREE(bkp_ctx->flog, SsFClose);
}

/*#***********************************************************************\
 *            bkp_log_dbase_begin
 *
 * 
 *
 * Parameters
 *    dbase     bkp_dbase_t* [in]
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */

static void bkp_log_dbase_begin(bkp_dbase_t* dbase)
{
        char buff[1024];

        if (!bkp_ctx->flog) {
            return;
        }

        bkp_log_curr_indent = 0;

        sprintf(buff, "\n%sDATABASE %s", bkp_log_indent[bkp_log_curr_indent],
            dbase->name);
        if (bkp_log_write(buff)) {
            goto FAILURE;
        }
        if (SsFFlush(bkp_ctx->flog)) {
            goto FAILURE;
        }

        return;

FAILURE:
        bkp_ctx->flog = NULL;
}

/*#***********************************************************************\
 *            bkp_log_dbase_end
 *
 * 
 *
 * Parameters
 *    dbase     bkp_dbase_t* [in]
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */

static void bkp_log_dbase_end(bkp_dbase_t* dbase)
{
        bkp_log_curr_indent = 0;
}

/*#***********************************************************************\
 *            bkp_log_table_begin
 *
 * 
 *
 * Parameters
 *    table     bkp_table_t* [in]
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_table_begin(bkp_table_t* table)
{
        char buff[1024];

        if (!bkp_ctx->flog) {
            return;
        }

        if (++bkp_log_curr_indent > BKP_MAX_LOG_INDENT) {
            bkp_log_curr_indent = BKP_MAX_LOG_INDENT;
        }

        sprintf(buff, "\n%sTABLE %s [%s]", bkp_log_indent[bkp_log_curr_indent],
            table->name,
            table->type == TABLE_TYPE_SOLID  ? "SOLID"  :
            table->type == TABLE_TYPE_MYISAM ? "MYISAM" : "<unknown>");

        if (bkp_log_write(buff)) {
            goto FAILURE;
        }
        if (SsFFlush(bkp_ctx->flog)) {
            goto FAILURE;
        }

        return;

FAILURE:
        bkp_ctx->flog = NULL;
}

/*#***********************************************************************\
 *            bkp_log_table_end
 *
 * 
 *
 * Parameters
 *    table     bkp_table_t* [in]
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_table_end(bkp_table_t* table)
{
        --bkp_log_curr_indent;
}

/*#***********************************************************************\
 *            bkp_log_trigger_begin
 *
 * 
 *
 * Parameters
 *    name     char* [in] Trigger name
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_trigger_begin(const char* name)
{
        char buff[1024];

        if (!bkp_ctx->flog) {
            return;
        }

        if (++bkp_log_curr_indent > BKP_MAX_LOG_INDENT) {
            bkp_log_curr_indent = BKP_MAX_LOG_INDENT;
        }

        sprintf(buff, "\n%sTRIGGER %s", bkp_log_indent[bkp_log_curr_indent], name);

        if (bkp_log_write(buff)) {
            goto FAILURE;
        }
        if (SsFFlush(bkp_ctx->flog)) {
            goto FAILURE;
        }

        return;

FAILURE:
        bkp_ctx->flog = NULL;
}

/*#***********************************************************************\
 *            bkp_log_trigger_end
 *
 * 
 *
 * Parameters
 *    name     char* [in] Trigger name
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_trigger_end(const char* name)
{
        --bkp_log_curr_indent;
}


/*#***********************************************************************\
 *            bkp_log_file_begin
 *
 * 
 *
 * Parameters
 *    name      char* [in]  File name
 *    extensio  char* [in]  File extension
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_file_begin(const char* name, const char* extension)
{
        char buff[1024];

        if (!bkp_ctx->flog) {
            return;
        }

        if (++bkp_log_curr_indent > BKP_MAX_LOG_INDENT) {
            bkp_log_curr_indent = BKP_MAX_LOG_INDENT;
        }

        sprintf(buff, "\n%sFILE %s.%s", bkp_log_indent[bkp_log_curr_indent],
            name, extension ? extension : "");
        if (bkp_log_write(buff)) {
            goto FAILURE;
        }
        if (SsFFlush(bkp_ctx->flog)) {
            goto FAILURE;
        }

        return;

FAILURE:
        bkp_ctx->flog = NULL;
}

/*#***********************************************************************\
 *            bkp_log_file_end
 *
 * 
 *
 * Parameters
 *    name      char* [in]  File name
 *    extensio  char* [in]  File extension
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_file_end(const char* name, const char* extension)
{
        --bkp_log_curr_indent;
}

/*#***********************************************************************\
 *            bkp_log_text
 *
 * 
 *
 * Parameters
 *    const char* text  [in] error text to write
 *
 * Return value
 *    none
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static void bkp_log_text(const char* text)
{
        if (!bkp_ctx->flog) {
            return;
        }

        if (bkp_log_write(text)) {
            goto FAILURE;
        }
        if (SsFFlush(bkp_ctx->flog)) {
            goto FAILURE;
        }

        return;

FAILURE:
        bkp_ctx->flog = NULL;
}

/*#***********************************************************************\
 *            bkp_log_write
 *
 * Write to the backup log
 *
 * Parameters
 *    const char* text  [in] text to write
 *
 * Return value
 *    int         zero on success, non-zero on failure
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
static int bkp_log_write(const char* text)
{
        size_t size = strlen(text);
        if (SsFWrite(text, 1, size, bkp_ctx->flog) != size) {
            return -1;
        }
        return 0;
}

/*#***********************************************************************\
 *            bkp_log_datetime
 *
 * Write current date-time in printable form to the buffer
 *
 * Parameters
 *    char*   buf     [out] buffer where date-time is to be written
 *    int     bufsize [in]  buffer size
 *
 * Return value
 *    none
 *
 * Globals used
 *    none
 */
static char* bkp_log_datetime(char* buff, int size)
{
       SsTimeT time = SsTime(NULL);
       SsCtime(&time, buff, size);

       for (size = strlen(buff); size--; ) {
          if (buff[size] != 0x0D && buff[size] != 0x0A) {
              break;
          }
          buff[size] = 0;
       }

       return buff;
}

#if MYSQL_VERSION_ID >= 50100
/*
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
W                                                                         W
W       I N F O R M A T I O N   S C H E M A   E X T E N S I O N           W
W                                                                         W
WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
*/
/************************************************************************
 *  INFORMATION_SCHEMA.SOLIDDB_BACKUP_STATUS table structure
 */
ST_FIELD_INFO soliddb_backup_status_fields[] =
{//   name              length      type           value flags  old_name            open_method
    {"STATUS",            8,    MYSQL_TYPE_STRING,    0,  0,    "Status",           SKIP_OPEN_TABLE},
    {"TOTAL SIZE",        4,    MYSQL_TYPE_LONGLONG,  0,  0,    "Total Size",       SKIP_OPEN_TABLE},
    {"DONE SIZE",         4,    MYSQL_TYPE_LONGLONG,  0,  0,    "Done Size",        SKIP_OPEN_TABLE},
    {"PERCENT COMPLETE",  2,    MYSQL_TYPE_SHORT,     0,  0,    "Percent Complete", SKIP_OPEN_TABLE},
    {"DIRECTORY",         120,  MYSQL_TYPE_STRING,    0,  0,    "Directory",        SKIP_OPEN_TABLE},
    {"MYISAM",            4,    MYSQL_TYPE_STRING,    0,  0,    "MyISAM",           SKIP_OPEN_TABLE},
    {"SYSTEM",            4,    MYSQL_TYPE_STRING,    0,  0,    "System",           SKIP_OPEN_TABLE},
    {"EMPTY DIR",         4,    MYSQL_TYPE_STRING,    0,  0,    "Empty Dir",        SKIP_OPEN_TABLE},
    {0,                   0,    MYSQL_TYPE_STRING,    0,  0,    0,                  0}
};

/*#***********************************************************************\
 *            soliddb_backup_status_fill
 *
 * Fill INFORMATION_SCHEMA.SOLIDDB_BACKUP_STATUS table
 * (used soliddb_init_backup_status() function)
 *
 * Parameters
 *    thd     MYSQL_THD   [in] MySQL thread
 *    tables  TABLE_LIST  [in] MySQL table
 *    cond    COND*       [in] not used
 *
 * Return value
 *    int     0 or error code
 *
 * Globals used
 *    bkp_ctx_t*  bkp_ctx
 */
int soliddb_backup_status_fill(MYSQL_THD thd, TABLE_LIST* tables, COND* cond)
{
        TABLE* table = tables->table;
        char*  status;
        double size_total;
        double size_done;
        uint32 percent;
        char*  directory;
        char*  opt_myisam;
        char*  opt_system;
        char*  opt_dempty;

        DBUG_ENTER("soliddb_sys_tables_fill");

        SsFlatMutexLock(bkp_ctx->mutex);
        status  = bkp_ctx->abort  ? (char*)"ABORTED" :
                  bkp_ctx->killed ? (char*)"KILLED"  :
                  bkp_ctx->failed ? (char*)"FAILED"  :
                  bkp_ctx->active ? (char*)"ACTIVE"  :
                                (char*)"FINISHED";
        size_total  = bkp_ctx->total_size;
        size_done   = bkp_ctx->done_size;
        directory   = bkp_ctx->dir;
        opt_myisam  = bkp_ctx->opt_myisam ? (char*)"ON" : (char*)"OFF";
        opt_system  = bkp_ctx->opt_system ? (char*)"ON" : (char*)"OFF";
        opt_dempty  = bkp_ctx->opt_dempty ? (char*)"ON" : (char*)"OFF";
        SsFlatMutexUnlock(bkp_ctx->mutex);

        percent = (uint32) ((size_total ? size_done / size_total : 0) * 100);

        restore_record(table, s->default_values);
        table->field[0]->store(status, strlen(status), system_charset_info);
        table->field[1]->store(size_total, TRUE);
        table->field[2]->store(size_done, TRUE);
        table->field[3]->store(percent, TRUE);
        table->field[4]->store(directory, strlen(directory), system_charset_info);
        table->field[5]->store(opt_myisam, strlen(opt_myisam), system_charset_info);
        table->field[6]->store(opt_system, strlen(opt_system), system_charset_info);
        table->field[7]->store(opt_dempty, strlen(opt_dempty), system_charset_info);

        if (schema_table_store_record(thd, table)) {
            DBUG_RETURN(1);
        }

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *            soliddb_init_backup_status
 *
 * Init INFORMATION_SCHEMA.SOLIDDB_BACKUP_STATUS table
 * (used in mysql_declare_plugin for this table)
 *
 * Parameters
 *    p       void* [in]  ST_SCHEMA_TABLE pointer
 *
 * Return value
 *    int     always 0
 *
 * Globals used
 *    none
 */
int soliddb_init_backup_status(void* p)
{
        DBUG_ENTER("soliddb_init_backup_status");
        ST_SCHEMA_TABLE* schema = (ST_SCHEMA_TABLE*) p;
        ss_dassert(schema != NULL);

        schema->fields_info = soliddb_backup_status_fields;
        schema->fill_table  = soliddb_backup_status_fill;

        DBUG_RETURN(0);
}

#endif /* MYSQL_VERSION_ID >= 50100 */
